#include "stdafx.h"

using namespace std;

#define AGILENT 0

/* FUNCTION *******************************************************************/
/**
@note Send
*******************************************************************************/
static ViStatus Send( ViSession vi, char *sCommand, bool bEnd )
{
  ViUInt32 len, count;
  len = static_cast<ViUInt32>(strlen(sCommand));
  ViStatus status;

  status = viSetAttribute( vi, VI_ATTR_SEND_END_EN, bEnd );
  if ( status < VI_SUCCESS ) return status;

  status = viWrite( vi, ViBuf(sCommand), len, &count );
  return status;
}

/* FUNCTION *******************************************************************/
/**
@note CheckResponse
*******************************************************************************/
static ViStatus CheckResponse(int lineNo, ViSession vi, char *sResponse )
{
  ViStatus status;
  ViUInt32 count;
  char buf[256];

  status = viRead( vi, (ViPBuf)buf, sizeof(buf)-1, &count );
  if ( !CompareResult( VI_SUCCESS, status, "CheckResponse", lineNo, __FILE__ ) )
  {
    return status;
  }
  buf[count] = '\0';
  
  if ( !CHECK( _stricmp( buf, sResponse ) == 0 ) )
  {
	  cout << "  Expected '" << sResponse << "'\n  Read '" << buf << "'" << endl;
	  cout << "  CheckResponse called from line " << lineNo << endl;
  }
  return status;
}

/* FUNCTION *******************************************************************/
/**
@note CheckNoError
*******************************************************************************/
static void CheckNoError( int lineNo, ViSession vi )
{
#if 0 //AGILENT
  // TRF: Agilent doesn't return 'AG' for VI_ATTR_MANF_ID
  if (! TR( CheckResponse( lineNo, vi, "+0,\"No error\"\n" ) ) )
	cout << "  CheckNoError called from line " << lineNo << endl;
#else
  ViUInt16 manfId;
  TR( viGetAttribute( vi, VI_ATTR_MANF_ID, &manfId ));
  if ( manfId == 'AG' )
  {
    if ( VI_SUCCESS != CheckResponse( lineNo, vi, "+0,\"No error\"\n" ) )
	    cout << "  CheckNoError called from line " << lineNo << endl;
  }
  else
  {
    if ( VI_SUCCESS != CheckResponse( lineNo, vi, "0,\"No error\"\n" ) )
	    cout << "  CheckNoError called from line " << lineNo << endl;
  }
#endif
}

/* FUNCTION *******************************************************************/
/**
@note CheckNoError
*******************************************************************************/
static void SleepOneSecond( ViSession vi )
{
#if 0 //AGILENT
  // TRF: Agilent doesn't return 'AG' for VI_ATTR_MANF_ID
    TR( Send( vi, "SLEEP 1", true ) );
#else
  ViUInt16 manfId;
  TR( viGetAttribute( vi, VI_ATTR_MANF_ID, &manfId ));
  if ( manfId == 'AG' )
  {
    TR( Send( vi, "SLEEP 1", true ) );
  }
  else if ( manfId == 'GK' ) // Tiny HislipServer from website
  {
    TR( Send( vi, "SLEEP 1000", true ) );
  }
  else if ( manfId == 'RS' )
  {
    TR( Send( vi, "test:scpi:internal:sync 1000", true ) );
  }
  else
  {
    cout << "Unknown server. Please enter a valid ManfId in line: " << __LINE__ << endl;
  }
#endif
}

/* FUNCTION *******************************************************************/
/**
@note CheckQueryError
*******************************************************************************/
static void CheckQueryError( int lineNo, ViSession vi )
{
  if ( VI_SUCCESS != CheckResponse( lineNo, vi, "-410,\"Query interrupted\"\n" ) )
	  cout << "  CheckQueryError called from line " << lineNo << endl;
}

/* FUNCTION *******************************************************************/
/**
@note TestMAV
*******************************************************************************/
static void TestMAV( ViSession vi, bool bIsSet, int lineNo )
{
  ViUInt16 stb;
  bool result;

  TR(viReadSTB( vi, &stb ));
  if ( bIsSet )
  {
    result = CHECK( ((stb & 0x10) > 0) );
  }
  else
  {
    result = CHECK( ((stb & 0x10) == 0) );
  }
  if (!result)
  {
	  cout << "  STB = " << stb << endl;
	  cout << "  TestMAV called from line " << lineNo << endl;
  }
}

/* FUNCTION *******************************************************************/
/**
@note CheckLockState
*******************************************************************************/
static void CheckLockState( int lineNo, ViSession vi, ViAttrState state )
{
  ViAttrState lockState = 0;
  TR( viGetAttribute( vi, VI_ATTR_RSRC_LOCK_STATE, &lockState ));
  if (! Check( lockState == state, "CheckLockState", lineNo, __FILE__ ) )
  {
	  cout << "Actual lockState " << lockState << " != expected state " << state << endl;
	  cout << "CheckLockState called from line " << lineNo << endl;
  }
}

#define TIMEOUT_MSEC 2000
/* FUNCTION *******************************************************************/
/**
@note TestHiSlipBasics
*******************************************************************************/
void TestHiSlipBasics( ViSession vi )
{
  char buf[256];
  ViUInt32 count;
  bool bDeviceSendsTriggerResponse = false;

  TR(viSetAttribute(vi,VI_ATTR_TMO_VALUE, TIMEOUT_MSEC ));
  TR(Send( vi, "*cls\n", true ));

  ViUInt16 manfId;
  TR( viGetAttribute( vi, VI_ATTR_MANF_ID, &manfId ));

  //////////////////////////////////////////////////////////////////////////////
  // First check if device sends a trigger response
  //////////////////////////////////////////////////////////////////////////////
  TR(viAssertTrigger( vi, VI_TRIG_PROT_DEFAULT ));
  if ( VI_SUCCESS == viRead( vi, ViPBuf(buf), sizeof(buf), VI_NULL ) )
  {
    bDeviceSendsTriggerResponse = true;
  }

  //////////////////////////////////////////////////////////////////////////////
  // Check Message Exchange Protocol 
  //////////////////////////////////////////////////////////////////////////////

  // Test normal response
  TR(Send( vi, "syst:err?\n", true )); CheckNoError(  __LINE__  , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // Test query interrupted. Must be detected only by server.
  TR(Send( vi, "syst:err?\nsyst:err?\n", false )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // Test query interrupted. Must be detected by server (and client)
  // (Here we test the cases "Fast client" or "Intermediate" 
  // (See also TestHiSlipTricky) 
  TR(Send( vi, "syst:err?\n", true ));
  TR(Send( vi, "syst:err?\n", true )); CheckQueryError(  __LINE__ , vi );

  // Test query interrupted with compound queries. Must be detected by server.
  TR(Send( vi, "syst:err?;err?\nsyst:err?\n", false )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );


  // Test query interrupted. 
  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  // Note! Client does not send RMT-delivered flag. 
  // Client has to delete stale 0,"No error" response with the next viWrite or viRead.
  TR(Send( vi, "syst:err?\n", true )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", true )); CheckNoError(  __LINE__  , vi );

  // Same without EOI
  TR(Send( vi, "syst:err?\n", false ));
  Sleep(1000);
  TR(Send( vi, "syst:err?\n", false )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // Read partial data
  TR(Send( vi, "syst:err?\n", false ));
  CR(VI_SUCCESS_MAX_CNT, viRead( vi, ViPBuf(buf), 10, VI_NULL ) );
  // Note! client did not read all data, but client has to delete 
  // pending response data with the next send command.
  TR(Send( vi, "syst:err?\n", false )); CheckQueryError(  __LINE__ , vi ); // TRF: this looks wrong: server never sees error
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // StatusByte and Remote/Local command does not interrupt a query
  TR(Send( vi, "syst:err?\n", false ));
  ViUInt16 stb;
  TR(viReadSTB( vi, &stb ));
  TR(viGpibControlREN( vi, VI_GPIB_REN_ASSERT_ADDRESS ));
  //CHECK((stb & 0x10) > 0); // MAV-Bit must be set.
  CheckNoError(  __LINE__  , vi );

  // Trigger interrupts a query
  TR(Send( vi, "syst:err?\n", false ));
  Sleep(1000); // TRF: Allow slow server to send response  
  TR(viAssertTrigger( vi, VI_TRIG_PROT_DEFAULT ));
  if ( bDeviceSendsTriggerResponse )
  {
    // We read the trigger response now.
    TR( viRead( vi, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  }
  else
  {
    // A read must run into a timeout.
    // Note. The client must delete incoming data with the next read command
    // since MessageIds does not correspond.
    CR(VI_ERROR_TMO, viRead( vi, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  }
  TR(Send( vi, "syst:err?\n", false )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );


  // Splitted commands. Agilent could have another behavior here, because
  // Agilent wants to send a response immediately.
  TR(Send( vi, "syst:err?;", false ));
  count = 10; // init count to a nonzero value
  CR(VI_ERROR_TMO, viRead( vi, ViPBuf(buf), sizeof(buf), &count ) );
  if ( manfId == 'AG' )
  {
	  // TRF: Agilent returns 13 char's of response without EOI
	  CHECK( count == 13 );
	  // Agilent does not return a response without the '?' at the end of *opc?
	  TR(Send( vi, "*opc?", true ));
	  CheckResponse(__LINE__, vi,"1\n");
  }
  else
  {
	  CHECK( count == 0 ); // R&S parser do not send a partial response here
	  TR(Send( vi, "*opc", true ));
	  CheckNoError(  __LINE__  , vi );  // Now we got the complete response.
  }


  //////////////////////////////////////////////////////////////////////////////
  // Test Status Byte
  //////////////////////////////////////////////////////////////////////////////
  TestMAV( vi, false, __LINE__ ); // MAV-Bit must be zero

  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  CR(VI_SUCCESS_MAX_CNT, viRead( vi, ViPBuf(buf), 5, VI_NULL ) );
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be still set
  TR( viRead( vi, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  TestMAV( vi, false, __LINE__ ); // All is read. MAV-Bit must be zero

  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  TR(Send( vi, "*SRE 0\n", true )); // Send a valid command that reset MAV
  TestMAV( vi, false, __LINE__ );
  TR(Send( vi, "syst:err?\n", false )); CheckQueryError(  __LINE__ , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // Same with viClear
  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  TR(viClear( vi ));
  TestMAV( vi, false, __LINE__ ); // Device Clear resets MAV
}

/* FUNCTION *******************************************************************/
/**
@note TestHiSlipOverlap
*******************************************************************************/
void TestHiSlipOverlap( ViSession vi )
{
  char buf[256];
  ViBoolean bOverlapEnabled = VI_FALSE;

  TR(viSetAttribute(vi,VI_ATTR_TCPIP_HISLIP_OVERLAP_EN, VI_TRUE ));
  if ( VI_SUCCESS != viGetAttribute( vi, VI_ATTR_TCPIP_HISLIP_OVERLAP_EN, &bOverlapEnabled ) ||
       !bOverlapEnabled )
  {
    cout << "Device does not support overlap mode. TestHiSlipOverlap() is skipped!\n";
    return;
  }

  TR(viSetAttribute(vi,VI_ATTR_TMO_VALUE, TIMEOUT_MSEC ));

  ViUInt16 manfId;
  TR( viGetAttribute( vi, VI_ATTR_MANF_ID, &manfId ));

  if ( manfId == 'RS' )
  {
    cout << "Don't worry! R&S server does not pass overlap mode test!\n";
  }

  TR(viClear( vi ));
  TR(Send( vi, "*cls\n", true ));

  //////////////////////////////////////////////////////////////////////////////
  // Check if Message Exchange Protocol is disabled correct.
  //////////////////////////////////////////////////////////////////////////////

  // Test normal response
  TR(Send( vi, "syst:err?\n", true )); CheckNoError(  __LINE__  , vi );
  TR(Send( vi, "syst:err?\n", false )); CheckNoError(  __LINE__  , vi );

  // Test overlapped response
  TR(Send( vi, "syst:err?\nsyst:err?\n", false ));
  TR(Send( vi, "syst:err?\n", false ));
  CheckNoError(  __LINE__  , vi );
  CheckNoError(  __LINE__  , vi );
  CheckNoError(  __LINE__  , vi );

  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  // Note! Client may not delete stale data here.
  TR(Send( vi, "syst:err?\n", true )); 
  CheckNoError(  __LINE__  , vi );
  CheckNoError(  __LINE__  , vi );

  // Same without EOI
  TR(Send( vi, "syst:err?\n", false ));
  Sleep(1000);
  TR(Send( vi, "syst:err?\n", false ));
  CheckNoError(  __LINE__  , vi );
  CheckNoError(  __LINE__  , vi );

  // Read partial data
  TR(Send( vi, "syst:err?\n", false ));
  CR(VI_SUCCESS_MAX_CNT, viRead( vi, ViPBuf(buf), 10, VI_NULL ) );
  // Note! client did not read all data, but client may not delete
  // pending response data with the next send command.
  TR(Send( vi, "syst:err?\n", false ));
  TR(viRead( vi, ViPBuf(buf), 10, VI_NULL ) );
  CheckNoError(  __LINE__  , vi );

  // StatusByte does not interrupt a query
  TR(Send( vi, "syst:err?\n", false ));
  ViUInt16 stb;
  TR(viReadSTB( vi, &stb ));
  CheckNoError(  __LINE__  , vi );

  // Trigger does not interrupt a query
  TR(Send( vi, "syst:err?\n", false ));
  TR(viAssertTrigger( vi, VI_TRIG_PROT_DEFAULT ));
  CheckNoError(  __LINE__  , vi );
  if ( manfId == 'AG' )
  {
	  // TRF: viAssertTrigger causes agilent server to send a '1' response...still not read
	  CheckResponse(__LINE__, vi, "1\n");
  }
  TestMAV( vi, false, __LINE__ ); // MAV-Bit must be zero

  //////////////////////////////////////////////////////////////////////////////
  // Test Status Byte
  //////////////////////////////////////////////////////////////////////////////
  TR( viClear(vi) );
  TestMAV( vi, false, __LINE__ ); // MAV-Bit must be zero

  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  CR(VI_SUCCESS_MAX_CNT, viRead( vi, ViPBuf(buf), 5, VI_NULL ) );
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be still set
  TR( viRead( vi, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  TestMAV( vi, false, __LINE__ ); // All is read. MAV-Bit must be zero

  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  TR(viAssertTrigger( vi, VI_TRIG_PROT_DEFAULT ));
  TestMAV( vi, true, __LINE__ ); // Trigger does not delete data.

  // Test with viClear
  TR(Send( vi, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( vi, true, __LINE__ ); // MAV-Bit should be set
  TR(viClear( vi ));
  // After viClear the MAV must be false immediately.
  TestMAV( vi, false, __LINE__ ); // Device Clear resets MAV

  TR( viSetAttribute( vi, VI_ATTR_TCPIP_HISLIP_OVERLAP_EN, VI_FALSE ));
}

/* FUNCTION *******************************************************************/
/**
@note TestHiSlipTricky
*******************************************************************************/
void TestHiSlipTricky( ViSession v1, ViSession v2 )
{
  char buf[256];
  ViUInt32 count;
  // Important: v1, v2 are currently opened. Do not send any data before the
  //            following tests. 

  // 1. Polling the status before sending any data is a very special case.
  // Please check if client is using the MessageID = 0xFFFFFF00-2.
  // Please check if server can manage this situation, because
  // the server does not know if this AsyncStatusQuery has overtaken a normal
  // data message.
  TestMAV( v1, false, __LINE__ );

  // Just reset event queue for safety
  TR(viSetAttribute(v1,VI_ATTR_TMO_VALUE, TIMEOUT_MSEC ));
  TR(Send( v1, "*cls\n", true ));

  // 2. Tricky test to simulate that a data message was overtaken by AsyncStatusQuery
  TR(Send( v1, "syst:err?\n", true )); // Prepare test and set the MAV bit
  Sleep(1000);
  TestMAV( v1, true, __LINE__ ); // MAV-Bit should be set now and a response is pending.
  
  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 1000,  NULL, NULL ) );

  TR(Send( v1, "syst:err?\n", true )); // Session v1 is blocked by server, but
  TestMAV( v1, false, __LINE__ ); // server should detect this situation by comparing the MessageIDs

  // 3. The client must discard the stale data and run into a timeout. 
  //    The client should ask with AsyncLockInfo the current lock state.
  CR(VI_ERROR_RSRC_LOCKED, viRead( v1, ViPBuf(buf), sizeof(buf), &count ) );
  CHECK( count == 0 ); 

  // 4. Unlock the server with client 2 and read the result with client 1.
  TR( viUnlock( v2 ) );
  CheckQueryError(  __LINE__ , v1 );
  TestMAV( v1, false, __LINE__ );

  // 5. Tricky test to simulate a fast client
  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 1000,  NULL, NULL ) );
  TR(Send( v1, "syst:err?\n", true ));
  TR(Send( v1, "syst:err?\n", true ));
  TestMAV( v1, false, __LINE__ );
  Sleep(1000);
  // Now both syst:err? queries must be blocked within TCP input buffer
  TR( viUnlock( v2 ) );
  CheckQueryError(  __LINE__ , v1 );

  TR( viLock( v1, VI_EXCLUSIVE_LOCK, 1000,  NULL, NULL ) );
  
  // 6.1 Check latency of locking 
  CheckLockState(  __LINE__  , v1, VI_EXCLUSIVE_LOCK );
  CheckLockState(  __LINE__  , v2, VI_EXCLUSIVE_LOCK );

  SleepOneSecond( v1 );
  SleepOneSecond( v1 );

  TR(Send( v1, "syst:err?\n", true ));
  // Note that syst:err? must be processed within the locked transaction!
  TR( viUnlock( v1 ) );
  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 4000,  NULL, NULL ) );
  // Even though the server is locked by another client, a response should be
  // available.
  CheckNoError(  __LINE__  , v1 );
  TR( viUnlock( v2 ) );

  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // 6.2 Check latency of locking
#if 0
  // TODO if locking issue is solved.  
#endif
  // Congratulation: If you reach this line without errors, you have done a very hard job :-)

  // TRF: Cleanup any lingering locks (these should error, ordinarily)
  CR( VI_ERROR_SESN_NLOCKED, viUnlock( v1 ) );
  CR( VI_ERROR_SESN_NLOCKED, viUnlock( v2 ) );
}

/* FUNCTION *******************************************************************/
/**
@note TestHiSlipSpecial
*******************************************************************************/
void TestHiSlipSpecial( ViSession v1, ViSession v2 )
{
  ViStatus status;

  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // Just reset event queue for safety
  TR(viSetAttribute(v1,VI_ATTR_TMO_VALUE, TIMEOUT_MSEC ));
  TR(Send( v1, "*cls\n", true ));
  TestMAV( v1, false, __LINE__ );

  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 1000,  NULL, NULL ) );
  // fill up input queue.
  
  for (;;)
  {
    status = Send( v1, "syst:err?\n", true );
    if ( VI_SUCCESS != status )
    {
      break;
    }
  }
  CHECK( status == VI_ERROR_RSRC_LOCKED );
  TR( viUnlock(v2) );
  CheckQueryError(  __LINE__ , v1 );
  TR(Send( v1, "*cls\n", true ));
  TR(Send( v1, "syst:err?\n", true )); CheckNoError(  __LINE__  , v1 );

  // Test timeout = 0
  TR(viSetAttribute(v1,VI_ATTR_TMO_VALUE, 0 ));
  TR(Send( v1, "syst:err?\n", true ));
  Sleep(1000); // response should be available.
  CheckNoError(  __LINE__  , v1 ); 
  
  SleepOneSecond( v1 );
  SleepOneSecond( v1 );
  SleepOneSecond( v1 );

  // A Device Clear transaction should at least wait 40 seconds before a fatal error
  // is detected.
  TR(viClear( v1 ));
  // A remote/local command should not run into a timeout.  
  TR(viGpibControlREN( v1, VI_GPIB_REN_ASSERT_ADDRESS ));

  // restore timeout
  TR(viSetAttribute(v1,VI_ATTR_TMO_VALUE, TIMEOUT_MSEC ));
}

#if 0 //AGILENT
//#define LOCK_FAILED_RESULT VI_ERROR_TMO
// TRF: avoid spurious viClear timeout
#define LOCK_TIMEOUT_MSEC 750
#else
//#define LOCK_FAILED_RESULT VI_ERROR_RSRC_LOCKED
#define LOCK_TIMEOUT_MSEC 1000
#endif

/* FUNCTION *******************************************************************/
/**
@note TestHiSlipLocking
*******************************************************************************/
void TestHiSlipLocking( ViSession v1, ViSession v2 )
{
  ViChar buf[256];
  //ViUInt32 count;

  ViUInt16 manfId;
  TR( viGetAttribute( v1, VI_ATTR_MANF_ID, &manfId ));

  ViChar key[256], keydummy[256];
  TR(viSetAttribute(v1, VI_ATTR_TMO_VALUE, LOCK_TIMEOUT_MSEC ));
  TR(viSetAttribute(v2, VI_ATTR_TMO_VALUE, LOCK_TIMEOUT_MSEC ));

#if 1
  TR(viClear( v1 ));
  TR(viClear( v2 ));

  // TRF Addition: should have no locks
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // Currently there is nothing to read.
  CR(VI_ERROR_TMO, viRead( v1, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  CR(VI_ERROR_TMO, viRead( v2, ViPBuf(buf), sizeof(buf), VI_NULL ) );

  // Test exclusive locked
  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 1000,  NULL, NULL ) );
  // TRF: I think VPP-3.4 says VI_ERROR_RSRC_LOCKED is for when a shared lock is
  // requested by an exclusive lock holder (RULE 3.6.12). Agilent returns VI_ERROR_TMO for these 4 tests.
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 100,  NULL, NULL ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 0,  NULL, key ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 100,  NULL, key ) );
  // There is nothing to read for v2
  CR(VI_ERROR_TMO, viRead( v2, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  // Data must be blocked, but Status Query and DeviceClear is allowed.
  TR(Send( v1, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( v1, false, __LINE__ ); 
  // There is nothing to read for v1, but error code should be VI_ERROR_RSRC_LOCKED!
  CR(VI_ERROR_RSRC_LOCKED, viRead( v1, ViPBuf(buf), sizeof(buf), VI_NULL ) );
  TR(viClear( v1 ));
  TR( viUnlock( v2 ) );
  // There is nothing to read for v1, but error code should be VI_ERROR_TMO now!
  CR(VI_ERROR_TMO, viRead( v1, ViPBuf(buf), sizeof(buf), VI_NULL ) );

  // TRF Addition: should have no locks
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // Test exclusive locked with no timeout
  TR( viLock( v2, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 100,  NULL, NULL ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 0,  NULL, key ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 100,  NULL, key ) );
  // Data must be blocked, but Status Query and DeviceClear is allowed.
  TR(Send( v1, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( v1, false, __LINE__ ); 
  TR(viClear( v1 ));
  TR( viUnlock( v2 ) );

  // TRF Addition: should have no locks
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // Test shared locked
  TR( viLock( v2, VI_SHARED_LOCK, 1000,  NULL, key ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 100,  NULL, NULL ) );
  // TRF: Do not pollute 'key' value needed below.
  TLF( viLock( v1, VI_SHARED_LOCK, 0,  NULL, keydummy ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 100,  NULL, keydummy ) );
  // Data must be blocked, but Status Query and DeviceClear is allowed.
  TR(Send( v1, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( v1, false, __LINE__ );
  TR(viClear( v1 ));

  // Same shared lock. Data is accepted
  TR( viLock( v1, VI_SHARED_LOCK, 0, key, keydummy ) );
  TR(Send( v1, "syst:err?\n", true ));
  CheckNoError(  __LINE__  , v1 );
  TR( viUnlock( v1 ) );
  TR( viLock( v1, VI_SHARED_LOCK, 100, key, keydummy ) );
  TR(Send( v1, "syst:err?\n", true ));
  CheckNoError(  __LINE__  , v1 );
  TR( viUnlock( v1 ) );
  TR( viUnlock( v2 ) );

  // TRF Addition: should have no locks
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

  // Test shared locked with no timeout
  TR( viLock( v2, VI_SHARED_LOCK, 0,  NULL, key ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  TLF( viLock( v1, VI_EXCLUSIVE_LOCK, 100,  NULL, NULL ) );
  // TRF: Do not pollute 'key' value needed below.
  TLF( viLock( v1, VI_SHARED_LOCK, 0,  NULL, keydummy ) );
  TLF( viLock( v1, VI_SHARED_LOCK, 100,  NULL, keydummy ) );
  // Data must be blocked, but Status Query and DeviceClear is allowed.
  TR(Send( v1, "syst:err?\n", true ));
  Sleep(1000);
  TestMAV( v1, false, __LINE__ ); 
  TR(viClear( v1 ));

  // Same shared lock. Data is accepted.
  TR( viLock( v1, VI_SHARED_LOCK, 0, key, keydummy ) );
  TR(Send( v1, "syst:err?\n", true ));
  CheckNoError(  __LINE__  , v1 );
  TR( viUnlock( v1 ) );
  TR( viLock( v1, VI_SHARED_LOCK, 100, key, keydummy ) );
  TR(Send( v1, "syst:err?\n", true ));
  CheckNoError(  __LINE__  , v1 );
  TR( viUnlock( v1 ) );
  TR( viUnlock( v2 ) );

  // Test Lock State
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );
  TR( viLock( v1, VI_SHARED_LOCK, 0, key, keydummy ) );
  CheckLockState(  __LINE__  , v1, VI_SHARED_LOCK );
  CheckLockState(  __LINE__  , v2, VI_SHARED_LOCK );
  // Get shared + exclusive lock for v1
  TR( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );
  CheckLockState(  __LINE__  , v1, VI_EXCLUSIVE_LOCK );
  CheckLockState(  __LINE__  , v2, VI_EXCLUSIVE_LOCK );
  // It is still allowed to get a shared lock for v2 while v1 has both locks
  TR( viLock( v2, VI_SHARED_LOCK, 0, key, keydummy ) );
  CheckLockState(  __LINE__  , v1, VI_EXCLUSIVE_LOCK );
  CheckLockState(  __LINE__  , v2, VI_EXCLUSIVE_LOCK );
  CR( VI_SUCCESS_NESTED_SHARED, viUnlock( v1 ) );
  CheckLockState(  __LINE__  , v1, VI_SHARED_LOCK );
  CheckLockState(  __LINE__  , v2, VI_SHARED_LOCK );
  TR( viUnlock( v1 ) );
  CheckLockState(  __LINE__  , v1, VI_SHARED_LOCK );
  CheckLockState(  __LINE__  , v2, VI_SHARED_LOCK );
  TR( viUnlock( v2 ) );
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );
#endif

  char *sFilename = "test.dat";
  switch ( manfId )
  {
     case 'RS': sFilename = "C:\\test.dat";
     // TODO: add your file
  }
  
  // Send: "mmem:data 'C:\\test.dat', #13aaa\n
  TR(viPrintf( v1, "mmem:data '%s',#13aaa\n", sFilename));
  TR(viPrintf( v1, "mmem:data? '%s'\n", sFilename));
  CheckResponse(__LINE__, v1,"#13aaa\n");


  TR( viLock( v2, VI_SHARED_LOCK, 0,  NULL, key ) );
  TR( viLock( v1, VI_SHARED_LOCK, 0, key, keydummy ) );
  TR( viLock( v1, VI_EXCLUSIVE_LOCK, 0,  NULL, NULL ) );

  // The following commands are locked
  SleepOneSecond( v1 );
  SleepOneSecond( v1 );
  SleepOneSecond( v1 );
  
  // Send: "mmem:data 'C:\\test.dat', #13bbb\n
  TR(viPrintf( v1, "mmem:data '%s',#13bbb\n", sFilename));

  CR( VI_SUCCESS_NESTED_SHARED, viUnlock( v1 ) );

  // The following command may not overtake the first command
  TR(viPrintf( v1, "mmem:data? '%s'\n", sFilename));
  // Use enough time to get the response.
  TR(viSetAttribute(v1, VI_ATTR_TMO_VALUE, 5000 ));
  CheckResponse(__LINE__, v1,"#13bbb\n");
  TR(viSetAttribute(v1, VI_ATTR_TMO_VALUE, LOCK_TIMEOUT_MSEC ));
 
  TR( viUnlock( v1 ) );
  TR( viUnlock( v2 ) );
  CheckLockState(  __LINE__  , v1, VI_NO_LOCK );
  CheckLockState(  __LINE__  , v2, VI_NO_LOCK );

}

/* FUNCTION *******************************************************************/
/**
@note TestVisaAttributes
*******************************************************************************/
void TestVisaAttributes( ViSession vi )
{
  // test new VISA attribute: VI_ATTR_MAX_MESSAGE_KB
  ViUInt32 size;
  TR( viGetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, &size ) );
  CHECK( size == 1024 );

  // TRF: reset event queue for safety (Agilent had a stale interrupted error here)
  TR(Send( vi, "*cls\n", true ));

  // Agilent VISA returns a parameter error for this case
  CR( VI_ERROR_INV_PARAMETER, viSetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, 0 ) );

  TR( viSetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, 1 ) );
  TR( viGetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, &size ) );
  CHECK( size == 1 );

  TR(Send( vi, "syst:err?\n", true )); // Test if connection still work
  CheckNoError(  __LINE__  , vi );

  TR( viSetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, 0xFFFF ) );
  TR( viGetAttribute( vi, VI_ATTR_TCPIP_HISLIP_MAX_MESSAGE_KB, &size ) );
  CHECK( size == 0xFFFF );

  TR(Send( vi, "syst:err?\n", true )); // Test if connection still work
  CheckNoError(  __LINE__  , vi );

  // test new VISA attribute: VI_ATTR_TCPIP_HISLIP_VERSION
  ViVersion version;
  TR( viGetAttribute( vi, VI_ATTR_TCPIP_HISLIP_VERSION, &version ) );
  CHECK( VI_VERSION_MAJOR(version) == 1 );
  CHECK( VI_VERSION_MINOR(version) == 0 );
  CHECK( VI_VERSION_SUBMINOR(version) == 0 );
  CR( VI_ERROR_ATTR_READONLY, viSetAttribute( vi, VI_ATTR_TCPIP_HISLIP_VERSION, version ) );

}

/* FUNCTION *******************************************************************/
/**
@note StartTest
*******************************************************************************/
void StartTest( ViRsrc name )
{
  cout << "Start test with resource: " << name << endl;
  ViSession defRM = NULL;
  ViSession v1 = NULL, v2 = NULL;

  TR( viOpenDefaultRM ( &defRM ) );

#if 1
  TR( viOpen( defRM, name, VI_NULL, 0, &v1 ) );
  cout << "1. Do basic tests\n";
  TestHiSlipBasics( v1 );
  cout << "2. Test new VISA attributes\n";
  TestVisaAttributes( v1 );
  cout << "3. Test overlap mode\n";
  // 
  TestHiSlipOverlap( v1 );
  TR( viClose( v1 ) );
#endif

  TR( viOpen( defRM, name, VI_NULL, 0, &v1 ) );
  TR( viOpen( defRM, name, VI_NULL, 0, &v2 ) );
  cout << "4. Do tricky tests\n";
  TestHiSlipTricky( v1, v2 );
  cout << "4. Do special tests\n";
  TestHiSlipSpecial( v1, v2 );
  
  cout << "5. Do locking tests\n";
  TestHiSlipLocking( v1, v2 );
  TR( viClose( v1 ) );
  TR( viClose( v2 ) );

  TR( viClose( defRM ) );
}

